/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "scene.h"

Scene* scene = 0;


Scene::Scene(): mpGraphics(NULL), currentGUI(-1) {
	initGLParameters();	
}

Scene::~Scene() {	
	clear();
}


void Scene::clear() {
	
	for(int i=0;i<display2DList.size();i++) {
		delete display2DList[i];
	}
	display2DList.clear();
		
	for(int i=0;i<display3DList.size();i++) {
		delete display3DList[i];
	}
	display3DList.clear();
	
	for(int i=0;i<guiLayers.size();i++) {
		delete guiLayers[i];
	}
	guiLayers.clear();	
	
	//printMeshesAndMaterialsNames ();	
	deleteMeshesAndMaterials ();	
}


void Scene::deleteMeshesAndMaterials() {
	for(std::set<Material*>::iterator it = materials.begin();it!=materials.end();it++) {
		delete (*it);
	}
	materials.clear();
	
	for(std::set<std::pair<Mesh*, int> >::iterator it = meshes.begin();it!=meshes.end();it++) {
		delete (*it).first;
	}
	meshes.clear();
}

void Scene::deleteUselessMeshesAndMaterials() {
	//TODO
}


void Scene::printMeshesAndMaterialsNames() {
	std::cout << "Names of the various materials: " << std::endl;
	for(std::set<Material*>::iterator it = materials.begin();it!=materials.end();it++) {
		std::cout << (*it)->getName() << " which if used in " << (*it)->getNbEntities() << " entities." << std::endl;
		std::cout << "material key: " << (*it)->getMaterialKey() << std::endl;
	}
	std::cout << "\n Names of the various meshes: " << std::endl;
	for(std::set<std::pair<Mesh*, int> >::iterator it = meshes.begin();it!=meshes.end();it++) {
		std::cout << (*it).first->getName() << " which if used in " << (*it).first->getNbEntities() << " entities, and has " << (*it).second << " textures." << std::endl;
	}		
}

GraphicOptions* Scene::getGraphicOptions() {
	return mpGraphics;
}

void Scene::setGraphicOptions(GraphicOptions *graphics) {
	mpGraphics = graphics;
}


void Scene::addDisplay2D(Display2D *disp2D) {
	display2DList.push_back(disp2D);
}

void Scene::addDisplay3D(Display3D *disp3D){
	display3DList.push_back(disp3D);
	disp3D->setRenderingMode(mpGraphics->getRenderingMode());
	disp3D->setMaxLightSources(mpGraphics->getMaxLightSources());
}

void Scene::addDisplay3D(Display3D *disp3D, renderingMode rendering) {
	display3DList.push_back(disp3D);
	disp3D->setRenderingMode(rendering);	
}

void Scene::addGuiLayer(Display3D *guiLayer) {
	guiLayers.push_back(guiLayer);
	guiLayer->setRenderingMode(mpGraphics->getRenderingMode());	
}


void Scene::addMaterialInTheSet(Material *material) {
	materials.insert(material);
}

void Scene::addMeshInTheSet(Mesh *mesh, int nbTextures) {
	meshes.insert(std::make_pair(mesh, nbTextures));
}


void Scene::initObjectsAndShaders () {
	
	/* init display environments*/
	for(int i=0;i<display3DList.size();i++) {
		display3DList[i]->setRenderingMode(mpGraphics->getRenderingMode());
		display3DList[i]->init(mpGraphics->getTextureMode());
	}
	for(int i=0;i<guiLayers.size();i++) {
		guiLayers[i]->setRenderingMode(mpGraphics->getRenderingMode());
		guiLayers[i]->init(mpGraphics->getTextureMode());
	}
			
	/* compile shaders	*/
	if(mpGraphics->isGLSLEnabled()) {
		for(int i=0;i<display3DList.size();i++) {
			display3DList[i]->compileShaders();
		}
		for(int i=0;i<guiLayers.size();i++) {
			guiLayers[i]->compileShaders();
		}		
	}
	
	/* fill the materials set and the meshes set */
	for(int i=0;i<display3DList.size();i++) {
		display3DList[i]->addMeshesInTheSet(meshes);
		display3DList[i]->addMaterialsInTheSet(materials);		
	}
	for(int i=0;i<guiLayers.size();i++) {
		guiLayers[i]->addMeshesInTheSet(meshes);
		guiLayers[i]->addMaterialsInTheSet(materials);		
	}	
	//printMeshesAndMaterialsNames();

	
	/* load meshes	*/
	for(std::set<std::pair<Mesh*, int> >::iterator it = meshes.begin();it!=meshes.end();it++) {
		//std::cout << ((*it).first)->getName() << std::endl;
		((*it).first)->initialize((*it).second);
	}
	
	//std::cout << "initialisation ok" << std::endl;
	
	
}


int Scene::getHit(int bufSize, int xMouse, int yMouse, float delX, float delY, float distMax) {

	int res = -1;	
	
	if(currentGUI <0) {
		int i = 0;
		
		while((i < display3DList.size()) && (res == -1)) {
			if(display3DList[i]->isInViewport(xMouse, yMouse)) {
				res = display3DList[i]->getHit(bufSize, xMouse, yMouse, delX, delY, distMax);
			}
			i++;
		}
		i=0;
		while((i < display2DList.size()) && (res == -1)) {
			res = display2DList[i]->getHit(bufSize, xMouse, yMouse, delX, delY);
			i++;
		}
		return res;
		
	} else {
		res = guiLayers[currentGUI]->getHit(bufSize, xMouse, yMouse, delX, delY, distMax);
		//std::cout << res << std::endl;
		return res;
	}
}

void Scene::render() {
	glClear (GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);  	
	
	for(int i=0;i<display3DList.size();i++) {
		display3DList[i]->render(mpGraphics);
	}
	
	glDisable(GL_DEPTH_TEST);
	for(int i=0;i<display2DList.size();i++) {
		display2DList[i]->render();
	}

	glEnable (GL_DEPTH_TEST);	
	
	if(currentGUI >=0) {
		glClear (GL_DEPTH_BUFFER_BIT);
		guiLayers[currentGUI]->render(mpGraphics);
	}
}


void Scene::initGLParameters() {
	glShadeModel(GL_SMOOTH);
	//glHint(GL_FOG_HINT, GL_NICEST);
	
	glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glFrontFace(GL_CCW);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
	//glDisable(GL_CULL_FACE);

}


void Scene::setClearColor(int numColor, float coeff) {
	Color color = Color(numColor, coeff, 0.);
	glClearColor(color.r, color.g, color.b, color.a);		
}

void Scene::setClearColor(Color color) {
	glClearColor(color.r, color.g, color.b, color.a);	
}

Display2D* Scene::getDisplay2D(int i) {
	if(i<display2DList.size()) {
		return display2DList[i];
	} else {
		return NULL;
	}
}

Display3D* Scene::getDisplay3D(int i){
	if(i<display3DList.size()) {
		return display3DList[i];
	} else {
		return NULL;
	}
}

Display3D* Scene::getGuiLayer(int i){
	if(i<guiLayers.size()) {
		return guiLayers[i];
	} else {
		return NULL;
	}
}

void Scene::setCurrentGUI(int i) {
	if(i<0) {
		currentGUI = -1;
	} else if(i<guiLayers.size()) {
		currentGUI = i;
	}
}

int Scene::getCurrentGUI() {
	return currentGUI;
}

