/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "shaderWriter.h"


ShaderWriter::ShaderWriter(): 
	mNbColorMaps(0), mNbNormalMaps(0), mNbGlossMaps(0), mNbEmiMaps(0),
	mNbLightDir(0), mNbLightSpot(0), mNbLightPoint(0) {
}
	
	
ShaderWriter::~ShaderWriter() {
}


void ShaderWriter::setNbMaps(int nbColorMaps, int nbNormalMaps, int nbGlossMaps, int nbEmiMaps) {
	mNbColorMaps = nbColorMaps;
	mNbNormalMaps = nbNormalMaps;
	mNbGlossMaps = nbGlossMaps;
	mNbEmiMaps = nbEmiMaps;
}
	
void ShaderWriter::setNbLights(int nbLightDir, int nbLightSpot, int nbLightPoint) {
	mNbLightDir = nbLightDir;
	mNbLightSpot = nbLightSpot;
	mNbLightPoint = nbLightPoint;
}



std::string ShaderWriter::writeLightingVertexShaderSourceCode() {
	std::string sourceCodeString;
	std::ostringstream sourceCode;
	
	int nbLights = mNbLightDir + mNbLightSpot + mNbLightPoint;
	
	sourceCode << "varying vec3 lightDir[" << nbLights << "]; \n" ;
	sourceCode << "varying float d[" << nbLights << "]; \n" ;
	sourceCode << "varying vec3 eyeDir; \n" ;
	
	if(mNbLightSpot) {
		sourceCode << "varying vec3 spotAxis[" << mNbLightSpot << "]; \n \n \n" ;
		
		sourceCode << "vec3 getSpotAxis( \n";
		sourceCode << "		in vec3 spotAxis, \n";
		sourceCode << "		in vec3 t, \n";
		sourceCode << "		in vec3 b, \n";
		sourceCode << "		in vec3 n) { \n \n";
		sourceCode << "	vec3 res; \n \n";
		sourceCode << "	res.x = dot(spotAxis, t); \n";
		sourceCode << "	res.y = dot(spotAxis, b); \n";
		sourceCode << "	res.z = dot(spotAxis, n); \n \n";
		sourceCode << "	return res; \n";
		sourceCode << "} \n";
	} else {
		sourceCode << "\n \n";
	}
	
	sourceCode << "void getDirAndDistLight( \n";
	sourceCode << "		in vec4 lightPosition, \n";
	sourceCode << "		in vec3 eyeD, \n";
	sourceCode << "		in vec3 t, \n";
	sourceCode << "		in vec3 b, \n";
	sourceCode << "		in vec3 n, \n";
	sourceCode << "		out vec3 lightDirection, \n";
	sourceCode << "		out float dist) { \n \n";

	sourceCode << "	vec3 v = lightPosition.xyz - eyeD; \n";

	sourceCode << "	lightDirection.x = dot(v, t); \n";
	sourceCode << "	lightDirection.y = dot(v, b); \n";
	sourceCode << "	lightDirection.z = dot(v, n); \n \n";

	sourceCode << "	dist = length(lightDirection); \n";
	sourceCode << "} \n \n";
	
	
	if(mNbLightDir) {
		sourceCode << "void getDirLightDir( \n";
		sourceCode << "		in vec3 direction, \n";
		sourceCode << "		in vec3 t, \n";
		sourceCode << "		in vec3 b, \n";
		sourceCode << "		in vec3 n, \n";
		sourceCode << "		out vec3 lightDirection) { \n \n";
		
		sourceCode << "	lightDirection.x = dot(direction, t); \n";
		sourceCode << "	lightDirection.y = dot(direction, b); \n";
		sourceCode << "	lightDirection.z = dot(direction, n); \n \n";
				
		sourceCode << "} \n \n";
	}
	
	sourceCode << "void fillDirAndDistArrays(in vec3 t, in vec3 b, in vec3 n) { \n";
	sourceCode << "	vec3 lightDirection; \n";
	sourceCode << "	float dist; \n";
	
	/* dir lights */
	for(int i=0;i<mNbLightDir;i++) {
		sourceCode << "	getDirLightDir (gl_LightSource[" << i << "].position.xyz, t, b, n, lightDirection); \n";
		sourceCode << "	lightDir[" << i << "] = lightDirection; \n";
		sourceCode << "	d[" << i << "] = 0.; \n";		
	}

	/* spot lights */
	for(int i=mNbLightDir;i<mNbLightDir+mNbLightSpot;i++) {
		sourceCode << "	getDirAndDistLight (gl_LightSource[" << i << "].position, eyeDir, t, b, n, lightDirection, dist); \n"; 
		sourceCode << "	lightDir[" << i << "] = lightDirection; \n";
		sourceCode << "	d[" << i << "] = dist; \n";
		sourceCode << "	spotAxis[" << i-mNbLightDir << "] = getSpotAxis(gl_LightSource[" << i << "].spotDirection, t, b, n); \n";
	}	
	
	/* point lights */
	for(int i=mNbLightDir+mNbLightSpot;i<nbLights;i++) {
		sourceCode << "	getDirAndDistLight (gl_LightSource[" << i << "].position, eyeDir, t, b, n, lightDirection, dist); \n"; 
		sourceCode << "	lightDir[" << i << "] = lightDirection; \n";
		sourceCode << "	d[" << i << "] = dist; \n";
	}
	sourceCode << "} \n";
		
	
	sourceCodeString = sourceCode.str();
	//std::cout << sourceCodeString << std::endl;
	
	return sourceCodeString;
}

	
std::string ShaderWriter::writeMaterialFragmentShaderSourceCode() {
	std::string sourceCode;
	
	if(mNbColorMaps) {
		sourceCode += "uniform sampler2D colorMap; \n";
	}
	if(mNbNormalMaps) {
		sourceCode += "uniform sampler2D normalMap; \n";
	}
	if(mNbGlossMaps) {
		sourceCode += "uniform sampler2D glossMap; \n";
	}	
	if(mNbEmiMaps) {
		sourceCode += "uniform sampler2D emiMap; \n";
	}
	
	sourceCode += "\n \n";
	
	sourceCode += "void material (in vec2 texCoord, out vec3 normal, out vec4 emiColor, out vec4 ambColor, out vec4 diffColor, out vec4 specColor) { \n" ;

	if(mNbNormalMaps) {
		sourceCode += "	normal = normalize(texture2D(normalMap, texCoord).xyz * 2.0 - 1.0); \n";
		sourceCode += "	normal.y = -normal.y; \n";
	} else {
		sourceCode += "	normal = vec3(0.,0.,1.); \n";		
	}	
	
	if(mNbEmiMaps) {
		sourceCode += "	emiColor = texture2D(emiMap,texCoord); \n";
	} else {
		sourceCode += "	emiColor = gl_FrontMaterial.emission; \n";
	}


	if(mNbColorMaps) {
		sourceCode += "	ambColor = texture2D(colorMap,texCoord); \n";
	} else {
		sourceCode += "	ambColor = gl_FrontMaterial.ambient; \n";
	}


	if(mNbColorMaps) {
		sourceCode += "	diffColor = texture2D(colorMap,texCoord); \n";
	} else {
		sourceCode += "	diffColor = gl_FrontMaterial.diffuse; \n";
	}	

	if(mNbGlossMaps) {
		sourceCode += "	specColor = texture2D(glossMap,texCoord); \n";
	} else {
		sourceCode += "	specColor.rgb = gl_FrontMaterial.specular.rgb; \n";
		sourceCode += "	specColor.a = gl_FrontMaterial.shininess; \n";		
	}	

	sourceCode += "} \n";
	
	return sourceCode;	
}
	

std::string ShaderWriter::writeLightingFragmentShaderSourceCode() {
	std::string sourceCodeString;
	std::ostringstream sourceCode;
	
	int nbLights = mNbLightDir + mNbLightSpot + mNbLightPoint;
	
	sourceCode << "varying vec3 lightDir[" << nbLights << "]; \n" ;
	sourceCode << "varying float d[" << nbLights << "]; \n" ;
	if(mNbLightSpot) {
		sourceCode << "varying vec3 spotAxis[" << mNbLightSpot << "]; \n" ;
	}
	sourceCode << "varying vec3 eyeDir; \n \n \n" ;
	
	
	sourceCode << "void lightingCoeff(in vec3 lightDir[" << nbLights << "], in float d[" << nbLights << "], \n";
	sourceCode << "		in vec3 vVec, in vec3 normal, in float shininess, \n";
	sourceCode << "		inout vec4 ambLighting, inout vec4 diffLighting, inout vec4 specLighting) { \n";
	sourceCode << "	vec3 lVec; \n";
	sourceCode << "	vec4 attAndDist; \n \n";

	if(mNbLightSpot) {
		sourceCode << "vec3 dirSpotAxis; \n" ;
	}	
	
	/* dir lights */
	for(int i=0;i<mNbLightDir;i++) {
		sourceCode << "	lVec = normalize(lightDir[" << i << "]); \n";	
		sourceCode << "	directionalLight(lVec, \n";
		sourceCode << "			gl_LightSource[" << i << "].ambient, gl_LightSource[" << i << "].diffuse, gl_LightSource[" << i << "].specular, \n";
		sourceCode << "			vVec, normal, shininess, \n";
		sourceCode << "			ambLighting, diffLighting, specLighting); \n";
	}
	
	/* spot lights */
	for(int i=mNbLightDir;i<mNbLightDir+mNbLightSpot;i++) {
		sourceCode << "	lVec = normalize(lightDir[" << i << "]); \n";	
		sourceCode << "	attAndDist = vec4 ( gl_LightSource[" << i << "]. constantAttenuation, \n";
		sourceCode << "			gl_LightSource[" << i << "]. linearAttenuation, \n";
		sourceCode << "			gl_LightSource[" << i << "]. quadraticAttenuation, \n";
		sourceCode << "			d[" << i << "]); \n";
		sourceCode << "	dirSpotAxis = normalize(spotAxis[" << i-mNbLightDir << "]); \n";	

		sourceCode << "	spotLight(attAndDist, lVec, dirSpotAxis, \n";
		sourceCode << "			gl_LightSource[" << i << "].ambient, gl_LightSource[" << i << "].diffuse, gl_LightSource[" << i << "].specular, \n";
		sourceCode << "			gl_LightSource[" << i << "].spotExponent, gl_LightSource[" << i << "].spotCosCutoff, \n";
		sourceCode << "			vVec, normal, shininess, \n";
		sourceCode << "			ambLighting, diffLighting, specLighting); \n";
	}	
	
	/* point lights */
	for(int i=mNbLightDir+mNbLightSpot;i<nbLights;i++) {
		sourceCode << "	lVec = normalize(lightDir[" << i << "]); \n";	
		sourceCode << "	attAndDist = vec4 ( gl_LightSource[" << i << "]. constantAttenuation, \n";
		sourceCode << "			gl_LightSource[" << i << "]. linearAttenuation, \n";
		sourceCode << "			gl_LightSource[" << i << "]. quadraticAttenuation, \n";
		sourceCode << "			d[" << i << "]); \n";

		sourceCode << "	pointLight(attAndDist, lVec, \n";
		sourceCode << "			gl_LightSource[" << i << "].ambient, gl_LightSource[" << i << "].diffuse, gl_LightSource[" << i << "].specular, \n";
		sourceCode << "			vVec, normal, shininess, \n";
		sourceCode << "			ambLighting, diffLighting, specLighting); \n";
	}	
	
	sourceCode << "} \n";
	
	sourceCodeString = sourceCode.str(); 
	//std::cout << sourceCodeString << std::endl;
	
	return sourceCodeString;	
}
	
